﻿using AutoMapper;
using Microsoft.EntityFrameworkCore;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Tsi1.BusinessLayer.Dtos;
using Tsi1.BusinessLayer.Helpers;
using Tsi1.BusinessLayer.Interfaces;
using Tsi1.DataLayer;
using Tsi1.DataLayer.Entities;

namespace Tsi1.BusinessLayer.Services
{
    public class ActivityService : IActivityService
    {
        private readonly Tsi1Context _context;
        private readonly IMapper _mapper;

        public ActivityService(Tsi1Context context, IMapper mapper)
        {
            _context = context;
            _mapper = mapper;
        }

        public async Task<ServiceResult<bool>> AttendVirtualClass(int activityId, int userId)
        {
            var result = new ServiceResult<bool>();

            var activity = await _context.Activities.FirstOrDefaultAsync(x => x.Id == activityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, activityId));
            }

            var attendance = await _context.Attendances
                .FirstOrDefaultAsync(x => x.ActivityId == activityId && x.UserId == userId);

            if (attendance == null)
            {
                attendance = new Attendance
                {
                    Activity = activity,
                    UserId = userId,
                    Date = DateTime.Now,
                };
            }
            
            return result;
        }

        public async Task<ServiceResult<int>> ActivityValidation(int activityId)
        {
            var result = new ServiceResult<int>();

            var activity = await _context.Activities.AsNoTracking().FirstOrDefaultAsync(x => x.Id == activityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, activityId));
            }

            return result;
        }

        private async Task<ServiceResult<User>> UserTypeValidation(int courseId, int userId)
        {
            var result = new ServiceResult<User>();

            var user = await _context.Users.AsNoTracking()
                .Include(x => x.UserType)
                .FirstOrDefaultAsync(x => x.Id == userId);

            if (user == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.UserDoesNotExist, userId));
                return result;
            }

            if (user.UserType.Name == UserTypes.Professor)
            {
                var isProfessorCourse = await _context.ProfessorCourses
                    .AsNoTracking()
                    .AnyAsync(x => x.ProfessorId == user.ProfessorId && x.CourseId == courseId);

                if (!isProfessorCourse)
                {
                    result.HasError = true;
                    result.AddMessage(string.Format(ErrorMessages.ProfessorCourseDoesNotExists, user.Id, courseId));

                }
            }

            if (user.UserType.Name == UserTypes.Student)
            {
                var isStudentCourse = await _context.StudentCourses
                    .AsNoTracking()
                    .AnyAsync(x => x.StudentId == user.StudentId && x.CourseId == courseId);

                if (!isStudentCourse)
                {
                    result.HasError = true;
                    result.AddMessage(string.Format(ErrorMessages.StudentCourseDoesNotExists, user.Id, courseId));

                }
            }

            result.Data = user;
            return result;
        }

        private async Task<ServiceResult<bool>> CourseValidation(int courseId)
        {
            var result = new ServiceResult<bool>();

            var course = await _context.Courses.AsNoTracking().FirstOrDefaultAsync(x => x.Id == courseId);

            if (course == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.CourseDoesNotExist, courseId));
            }

            return result;
        }

        public async Task<ServiceResult<List<ActivityDto>>> GetAll(int courseId)
        {
            var result = new ServiceResult<List<ActivityDto>>();

            var validation = await this.CourseValidation(courseId);

            if (validation.HasError)
            {
                result.HasError = true;
                result.AddMessage(validation.Message);
                return result;
            }

            var activities = await _context.Activities.AsNoTracking().Where(x => x.CourseId == courseId).ToListAsync();

            result.Data = _mapper.Map<List<ActivityDto>>(activities);
            return result;
        }

        public async Task<ServiceResult<int>> Create(ActivityCreateDto newActivity, int userId)
        {
            var result = new ServiceResult<int>();

            var courseValidation = await this.CourseValidation(newActivity.CourseId);

            if (courseValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(courseValidation.Message);
                return result;
            }

            var userTypeValidation = await this.UserTypeValidation(newActivity.CourseId, userId);

            if (userTypeValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userTypeValidation.Message);
                return result;
            }

            var activity = _mapper.Map<Activity>(newActivity);
            _context.Activities.Add(activity);

            await _context.SaveChangesAsync();

            result.Data = activity.Id;
            return result;
        }

        public async Task<ServiceResult<int>> Modify(int activityId, ActivityModifyDto activityDto, int userId)
        {
            var result = new ServiceResult<int>();

            var activity = await _context.Activities.FirstOrDefaultAsync(x => x.Id == activityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, activityId));
                return result;
            }

            var userTypeValidation = await this.UserTypeValidation(activity.CourseId, userId);

            if (userTypeValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userTypeValidation.Message);
                return result;
            }

            _mapper.Map(activityDto, activity);

            await _context.SaveChangesAsync();

            result.Data = activity.Id;
            return result;
        }

        public async Task<ServiceResult<int>> Delete(int activityId, int userId)
        {
            var result = new ServiceResult<int>();

            var activity = await _context.Activities.FirstOrDefaultAsync(x => x.Id == activityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, activityId));
                return result;
            }

            var userTypeValidation = await this.UserTypeValidation(activity.CourseId, userId);

            if (userTypeValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userTypeValidation.Message);
                return result;
            }

            _context.Activities.Remove(activity);

            await _context.SaveChangesAsync();

            result.Data = activity.Id;
            return result;
        }

        public async Task<ServiceResult<GradeDto>> AddOrModifyGrade(GradeDto gradeDto, int userId)
        {
            var result = new ServiceResult<GradeDto>();

            var activity = await _context.Activities.FirstOrDefaultAsync(x => x.Id == gradeDto.ActivityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, gradeDto.ActivityId));
                return result;
            }

            var userTypeValidation = await this.UserTypeValidation(activity.CourseId, userId);

            if (userTypeValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userTypeValidation.Message);
                return result;
            }

            var userStudent = await _context.Users
                .Include(x => x.Student)
                .FirstOrDefaultAsync(x => x.Id == gradeDto.UserId);

            if (userStudent == null || userStudent.Student == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.StudentDoesNotExist, gradeDto.UserId));
                return result;
            }

            var userStudentValidation = await this.UserTypeValidation(activity.CourseId, userStudent.Id);

            if (userStudentValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userStudentValidation.Message);
                return result;
            }

            var grade = await _context.Grades
                .FirstOrDefaultAsync(x => x.ActivityId == gradeDto.ActivityId && x.StudentId == userStudent.StudentId);

            if (grade == null)
            {
                grade = new Grade
                {
                    ActivityId = gradeDto.ActivityId,
                    StudentId = (int)userStudent.StudentId
                };

                _context.Grades.Add(grade);
            }
            
            grade.Value = gradeDto.Value;

            
            await _context.SaveChangesAsync();

            result.Data = gradeDto;
            return result;
        }

        public async Task<ServiceResult<int>> DeleteGrade(int myUserId, int activityId, int userId)
        {
            var result = new ServiceResult<int>();

            var userStudent = await _context.Users
                .AsNoTracking()
                .Include(x => x.Student)
                .FirstOrDefaultAsync(x => x.Id == userId);

            if (userStudent == null || userStudent.Student == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.StudentDoesNotExist, userStudent.Id));
                return result;
            }

            var grade = await _context.Grades
                .FirstOrDefaultAsync(x => x.ActivityId == activityId && x.StudentId == userStudent.StudentId);

            if (grade == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.GradeDoesNotExist, activityId, userId));
                return result;
            }

            var activity = await _context.Activities.FirstOrDefaultAsync(x => x.Id == activityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, activityId));
                return result;
            }

            var userTypeValidation = await this.UserTypeValidation(activity.CourseId, userId);

            if (userTypeValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userTypeValidation.Message);
                return result;
            }

            _context.Grades.Remove(grade);
            await _context.SaveChangesAsync();

            return result;
        }

        public async Task<ServiceResult<List<GradeDto>>> GetMyGrades(int activityId, int userId)
        {
            var result = new ServiceResult<List<GradeDto>>();

            var activity = await _context.Activities.AsNoTracking().FirstOrDefaultAsync(x => x.Id == activityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, activityId));
                return result;
            }

            var userTypeValidation = await this.UserTypeValidation(activity.CourseId, userId);

            if (userTypeValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userTypeValidation.Message);
                return result;
            }

            var user = userTypeValidation.Data;

            var grades = await _context.Grades
                .AsNoTracking()
                .Where(x => x.ActivityId == activityId && x.StudentId == user.StudentId)
                .ToListAsync();

            result.Data = _mapper.Map<List<GradeDto>>(grades);
            return result;
        }

        public async Task<ServiceResult<List<GradeDto>>> GetAllGrades(int activityId, int userId)
        {
            var result = new ServiceResult<List<GradeDto>>();

            var activity = await _context.Activities.AsNoTracking().FirstOrDefaultAsync(x => x.Id == activityId);

            if (activity == null)
            {
                result.HasError = true;
                result.AddMessage(string.Format(ErrorMessages.ActivityDoesNotExist, activityId));
                return result;
            }

            var userTypeValidation = await this.UserTypeValidation(activity.CourseId, userId);

            if (userTypeValidation.HasError)
            {
                result.HasError = true;
                result.AddMessage(userTypeValidation.Message);
                return result;
            }

            var user = userTypeValidation.Data;

            var grades = await _context.Grades
                .AsNoTracking()
                .Where(x => x.ActivityId == activityId)
                .ToListAsync();

            var studentIds = grades.Select(x => x.StudentId).ToList();

            var users = await _context.Users
                .AsNoTracking()
                .Where(x => studentIds.Contains((int)x.StudentId))
                .ToListAsync();

            var gradeDtos = new List<GradeDto>();

            foreach (var grade in grades)
            {
                var userStudent = users.FirstOrDefault(x => x.StudentId == grade.StudentId);

                var gradeDto = new GradeDto
                {
                    ActivityId = grade.ActivityId,
                    UserId = userStudent.Id,
                    Value = grade.Value
                };

                gradeDtos.Add(gradeDto);
            }

            result.Data = gradeDtos;
            return result;
        }
    }
}
